import asyncio
import math
import time
from binascii import unhexlify

import serial
from loguru import logger

from ..interfaces import database
from ..utils import data_formatter, config, bluetooth


class Parameter:
    def __init__(self, data, field, start_byte, stop_byte, byte_order='big', signed=False, precision=1.0):
        self.field = field
        self.start_byte = start_byte
        self.stop_byte = stop_byte
        self.byte_order = byte_order
        self.signed = signed
        self.precision = precision
        self.data = data[start_byte: stop_byte]
        self.value = int.from_bytes(data[self.start_byte: self.stop_byte], byteorder=self.byte_order,
                                    signed=self.signed) * self.precision
        if self.precision < 1:
            self.value = round(self.value, math.ceil(math.log(1 / self.precision, 10)))

    def get_bits(self, index=None):
        string_bytes = self.data
        bytes_list = []
        for i in range(len(string_bytes)):
            bytes_list.append(string_bytes[i:i + 1])

        bits = []
        for byte in bytes_list:
            for i in range(8):
                bits.append((ord(byte) >> i) & 1)
        return bits


def initialize():
    try:
        ser = serial.Serial(
            port=config.homesolar_config['BLUETOOTH']['port'],
            baudrate=9600,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS)

        try:
            ser.open()
            time.sleep(1)
        except Exception as e:
            ser.close()
            logger.warning(f"Something went wrong when trying to open Bluetooth as a Serial Connection [{e}]")
            initialize()

        first_boot = True
        while True:
            try:
                ser.write(unhexlify(config.homesolar_config['BLUETOOTH']['request_code']))
                if first_boot:
                    time.sleep(1)
                    first_boot = False
                antw33 = ser.read(140)

                params = [Parameter(antw33, "Voltage", 4, 6, 'big', True, 0.1),
                          Parameter(antw33, "Current", 70, 74, 'big', True, 0.1),
                          Parameter(antw33, "Power", 111, 115, 'big', True),
                          Parameter(antw33, "Capacity", 75, 79, 'big', False, 0.000001),
                          Parameter(antw33, "ChargeStatus", 103, 104), Parameter(antw33, "DischargeStatus", 104, 105),
                          Parameter(antw33, "BalanceStatus", 105, 106), Parameter(antw33, "BalanceCells", 133, 136),
                          Parameter(antw33, "Temp1", 91, 93), Parameter(antw33, "Temp2", 93, 95),
                          Parameter(antw33, "Temp3", 95, 97), Parameter(antw33, "Temp4", 97, 99),
                          Parameter(antw33, "Temp5", 99, 101), Parameter(antw33, "Temp6", 101, 103),
                          Parameter(antw33, "MinCell", 118, 119), Parameter(antw33, "MaxCell", 115, 116),
                          Parameter(antw33, "MinVolt", 119, 121, 'big', False, 0.001),
                          Parameter(antw33, "MaxVolt", 116, 118, 'big', False, 0.001),
                          Parameter(antw33, "AvgVolt", 121, 123, precision=0.001),
                          Parameter(antw33, "Cell1", 6, 8, 'big', False, 0.001),
                          Parameter(antw33, "Cell2", 8, 10, 'big', False, 0.001),
                          Parameter(antw33, "Cell3", 10, 12, 'big', False, 0.001),
                          Parameter(antw33, "Cell4", 12, 14, 'big', False, 0.001),
                          Parameter(antw33, "Cell5", 14, 16, 'big', False, 0.001),
                          Parameter(antw33, "Cell6", 16, 18, 'big', False, 0.001),
                          Parameter(antw33, "Cell7", 18, 20, 'big', False, 0.001),
                          Parameter(antw33, "Cell8", 20, 22, 'big', False, 0.001),
                          Parameter(antw33, "Cell9", 22, 24, 'big', False, 0.001),
                          Parameter(antw33, "Cell10", 24, 26, 'big', False, 0.001),
                          Parameter(antw33, "Cell11", 26, 28, 'big', False, 0.001),
                          Parameter(antw33, "Cell12", 28, 30, 'big', False, 0.001),
                          Parameter(antw33, "Cell13", 30, 32, 'big', False, 0.001),
                          Parameter(antw33, "Cell14", 32, 34, 'big', False, 0.001),
                          Parameter(antw33, "Cell15", 34, 36, 'big', False, 0.001),
                          Parameter(antw33, "Cell16", 36, 38, 'big', False, 0.001),
                          Parameter(antw33, "Cell17", 38, 40, 'big', False, 0.001),
                          Parameter(antw33, "Cell18", 40, 42, 'big', False, 0.001),
                          Parameter(antw33, "Cell19", 42, 44, 'big', False, 0.001),
                          Parameter(antw33, "Cell20", 44, 46, 'big', False, 0.001),
                          Parameter(antw33, "Cell21", 46, 48, 'big', False, 0.001),
                          Parameter(antw33, "Cell22", 48, 50, 'big', False, 0.001),
                          Parameter(antw33, "Cell23", 50, 52, 'big', False, 0.001),
                          Parameter(antw33, "Cell24", 52, 54, 'big', False, 0.001)]

                voltage = 0
                fields = {}
                for data in params:
                    fields[data.field] = data.value

                    if data.field == 'AvgVolt':
                        voltage = data.value

                    if data.field == 'BalanceCells':
                        for index, bal in enumerate(data.get_bits()):
                            fields[f"Bal{index}"] = bal

                    if data.field == 'AvgVolt':
                        fields['SoC'] = bluetooth.ManualSoC().get_soc(data.value)

                if voltage == 0:
                    ser.close()
                    time.sleep(1)
                    ser.open()
                    time.sleep(1)

                data = {
                    "measurement": "Antw33-BMS",
                    "fields": data_formatter.flatten_dict(fields)
                }
                logger.debug(data)
                asyncio.run(database.write_sensor_data(data))

            except:
                ser.close()
                time.sleep(1)
                ser.open()
                time.sleep(1)

    except Exception as ex:
        print(ex)
    finally:
        try:
            ser.close()
        except Exception as e:
            logger.warning(f"Something went wrong when trying to close Bluetooth as Serial Connection [{e}]")



import asyncio

from loguru import logger
from paho.mqtt import client as paho_mqtt

# project's imports
from ..interfaces import database
from ..utils import config
from ..utils.mqtt import MqttTopic

mqtt_client = None
connected_clients = []


def initialize():
    try:
        global mqtt_client
        mqtt_client = paho_mqtt.Client()

        host = config.homesolar_config['MQTT']['host']
        port = config.homesolar_config['MQTT']['port']
        username = config.homesolar_config['MQTT']['username']
        password = config.homesolar_config['MQTT']['password']
        keepalive = config.homesolar_config['MQTT']['keepalive']
        reconnect = config.homesolar_config['MQTT']['reconnect_delay']

        mqtt_client.reconnect_delay_set(max_delay=reconnect)
        mqtt_client.on_message = on_message
        mqtt_client.on_connect = on_connect
        mqtt_client.username_pw_set(username, password)
        mqtt_client.connect(host, port, keepalive)
        mqtt_client.loop_forever()
    except Exception as e:
        logger.exception(f"Something went wrong when initialize client: {e}")


def on_message(current_client: paho_mqtt.Client, userdata, msg):
    # Handles messages on Sensors Topic
    if MqttTopic.SENSORS[:-2] in str(msg.topic):
        sensor_name = str(msg.topic).replace(MqttTopic.SENSORS[:-2] + "/", "")
        data = {
            "name": sensor_name,
            "payload": msg.payload
        }

        asyncio.run(database.write_sensor_data(data))

    # Handles messages on Request Topic
    elif MqttTopic.REQUEST[:-2] in str(msg.topic):
        pass

    # Handles messages on Client Topic
    elif MqttTopic.CLIENT[:-2] in str(msg.topic):
        logger.debug("Client updated")
        client_id = str(msg.topic).replace(MqttTopic.CLIENT[:-2] + "/", "")
        status = str(msg.payload, 'utf-8')
        try:
            if status == "Connected":
                connected_clients.append(client_id)
            else:
                connected_clients.remove(client_id)
        except Exception as e:
            logger.warning(f"Something went wrong when processing clients connection {e}")


def on_connect(current_client: paho_mqtt.Client, userdata, flags, reconnect):
    try:
        current_client.subscribe(MqttTopic.SENSORS)
        current_client.subscribe(MqttTopic.REQUEST)
        current_client.subscribe(MqttTopic.CLIENT)
        message = "Client reconnected successfully" if reconnect else "Client connected successfully"
        logger.info(message)
    except Exception as e:
        logger.exception(f"Something went wrong while trying to subscribes {e}")



import asyncio
import json

from loguru import logger

# project's imports
from ..services import influxdb, sqlite
from ..utils import datetime, config, data_formatter
from ..utils import influxdb as influx_utils
from ..utils import sqlite as sqlite_utils
from ..utils.sqlite import SensorData


# Used on All database
async def write_sensor_data(data):
    sensor_data = data_formatter.format_sensor_data(data)
    await influxdb.write(sensor_data)
    await sqlite.write(sqlite_utils.generate_add_sensor_data_sql(sensor_data))
    logger.debug(json.dumps(sensor_data))


# Used only on InfluxDB
def get_battery_charge(date):
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_charge_measurement'],
        config.homesolar_config['DATA']['battery_charge_field'],
        datetime.stringify_timestamp(date),
        datetime.get_next_day(date), "DAY"
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result)


def get_battery_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_power_measurement'],
        config.homesolar_config['DATA']['battery_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result, timescale)


def get_solar_production(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['solar_production_measurement'],
        config.homesolar_config['DATA']['solar_production_field'],
        start_time,
        stop_time,
        timescale
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result, timescale)


def get_grid_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['grid_power_measurement'],
        config.homesolar_config['DATA']['grid_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result, timescale)


def get_inverter_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['inverter_power_measurement'],
        config.homesolar_config['DATA']['inverter_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result, timescale)


def get_home_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_combined_tables_flux(
        [config.homesolar_config['DATA']['grid_power_measurement'],
         config.homesolar_config['DATA']['grid_power_field']],
        [config.homesolar_config['DATA']['inverter_power_measurement'],
         config.homesolar_config['DATA']['inverter_power_field']],
        start_time,
        stop_time,
        timescale
    )
    result = asyncio.run(influxdb.query(flux))
    logger.debug(result)
    return influx_utils.serialize(result, timescale)


# Used only on Sqlite
def get_sensor_data(measurement, field):
    sql = sqlite_utils.generate_get_sensor_data_sql(measurement, field)
    result = asyncio.run(sqlite.execute(sql))

    sensor_data: SensorData = result.scalar()
    if sensor_data is None:
        logger.warning("No sensor data with specified name found")
        return None
    else:
        return data_formatter.parse_to_float_if_possible(sensor_data.value)



from loguru import logger
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from ..utils import sqlite

sqlite_engine = create_async_engine("sqlite+aiosqlite:///homesolar.db", echo=False)
sqlite_session = sessionmaker(bind=sqlite_engine, expire_on_commit=False, class_=AsyncSession)


async def reinitialize_tables():
    async with sqlite_engine.begin() as conn:
        await conn.run_sync(sqlite.Base.metadata.drop_all)
        await conn.run_sync(sqlite.Base.metadata.create_all)


async def write(data):
    try:
        async with sqlite_session() as session:
            async with session.begin():

                if type(data) is list:
                    for stmt in data:
                        await session.execute(stmt)
                else:
                    await session.execute(data)
            await session.commit()
            await session.close()

        logger.info("Data saved successfully!")
    except Exception as e:
        logger.exception(f"Data not saved! [{e}]")


async def execute(statement):
    result = None
    try:
        async with sqlite_session() as session:
            async with session.begin():
                result = await session.execute(statement)
            await session.commit()
            await session.close()
    except Exception as e:
        logger.exception(f"Something went wrong when executing an sql! [{e}]")
    finally:
        return result
